package ru.ifmo.rain.abdrashitov.java.lang;

import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;
import java.util.Comparator;
import java.util.Formatter;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public final class QuickString implements java.io.Serializable, Comparable<QuickString>,
		CharSequence {
	
	private static final long serialVersionUID = -6849794470754667710L;
	private static final QuickString EMPTY = new QuickString();
	private static int rnd = 0;
	
	private String value;
	private int count;
	private int hash;
	private QuickString left;
	private QuickString right;
		
    public QuickString() {
    	count = 0;
    }

    public QuickString(String original) {
    	this.value = original;
    	this.count = value.length();
    	this.left = EMPTY;
    	this.right = EMPTY;
    }
    
    public QuickString(char value[]) {
    	this(value, 0, value.length);
    }

    public QuickString(char value[], int offset, int count) {
    	this(new String(value, offset, count));
	}
    
    private QuickString(String value, QuickString left, QuickString right) {
    	this.value = value;
    	this.count = left.count + value.length() + right.count;
    	this.left = left;
    	this.right = right;
    }

    public QuickString(int[] codePoints, int offset, int count) {
    	this(new String(codePoints, offset, count));
	}

    @Deprecated
    public QuickString(byte ascii[], int hibyte, int offset, int count) {
    	this(new String(ascii, hibyte, offset, count));
	}

    @Deprecated
    public QuickString(byte ascii[], int hibyte) {
        this(ascii, hibyte, 0, ascii.length);
    }

    public QuickString(byte bytes[], int offset, int length, QuickString charsetName)
			throws UnsupportedEncodingException {
    	this(new String(bytes, offset, length, charsetName.toString()));
	}

    public QuickString(byte bytes[], int offset, int length, Charset charset) {
    	this(new String(bytes, offset, length, charset));
	}

    public QuickString(byte bytes[], QuickString charsetName)
			throws UnsupportedEncodingException {
		this(bytes, 0, bytes.length, charsetName);
	}

    public QuickString(byte bytes[], Charset charset) {
		this(bytes, 0, bytes.length, charset);
	}

    public QuickString(byte bytes[], int offset, int length) {
    	this(new String(bytes, offset, length));
	}

    public QuickString(byte bytes[]) {
		this(bytes, 0, bytes.length);
	}
    
    public QuickString(StringBuffer buffer) {
    	this(buffer.toString());
    }

    public QuickString(StringBuilder builder) {
    	this(builder.toString());
    }

	QuickString(int offset, int count, char value[]) {
		this(new String(value, offset, count));
	}

    public int length() {
        return count;
    }

    public boolean isEmpty() {
    	return count == 0;
    }

    public char charAt(int index) {
        if ((index < 0) || (index >= count)) {
            throw new StringIndexOutOfBoundsException(index);
        }
        return getChar(index);
    }
    
    private char getChar(int i) {
    	if (i < left.count) {
    		return left.getChar(i);
    	} else if (i < left.count + value.length()) {
    		return value.charAt(i - left.count);
    	} else {
    		return right.getChar(i - left.count - value.length());
    	}
    }
    
    public int codePointAt(int index) {
    	QuickString str = substring(index, Math.min(index + 2, count));
    	return str.toString().codePointAt(0);
	}

    public int codePointBefore(int index) {
    	QuickString str = substring(Math.max(0, index - 2), index);
    	return str.toString().codePointBefore(str.count);
	}

    public int codePointCount(int beginIndex, int endIndex) {
    	QuickString str = substring(beginIndex, endIndex);
    	return str.toString().codePointCount(0, str.count);
	}
    
    public int offsetByCodePoints(int index, int codePointOffset) {
    	return toString().offsetByCodePoints(index, codePointOffset);
	}

    void getChars(char dst[], int dstBegin) {
    	if (count == 0) {
    		return;
    	}
    	left.getChars(dst, dstBegin);
    	value.getChars(0, value.length(), dst, dstBegin + left.count);
    	right.getChars(dst, dstBegin + left.count + value.length());
    }

    public void getChars(int srcBegin, int srcEnd, char dst[], int dstBegin) {
    	substring(srcBegin, srcEnd).getChars(dst, dstBegin);
    }

    @Deprecated
    public void getBytes(int srcBegin, int srcEnd, byte dst[], int dstBegin) {
    	QuickString str = substring(srcBegin, srcEnd);
    	str.toString().getBytes(0, str.count, dst, dstBegin);
    }

    public byte[] getBytes(QuickString charsetName)
			throws UnsupportedEncodingException {
    	return toString().getBytes(charsetName.toString());
	}

    public byte[] getBytes(Charset charset) {
    	return toString().getBytes(charset);
	}

    public byte[] getBytes() {
    	return toString().getBytes();
	}

    public boolean equals(Object anObject) {
		if (this == anObject) {
			return true;
		}
		if (anObject instanceof QuickString) {
			QuickString other = (QuickString) anObject;
			if (this.count == other.count) {
				for (int i = 0; i < count; i++) {
					if (this.getChar(i) != other.getChar(i)) {
						return false;
					}
				}
				return true;
			}
		}
		return false;
	}

    public boolean contentEquals(StringBuffer sb) {
		synchronized (sb) {
			return contentEquals((CharSequence) sb);
		}
	}

    public boolean contentEquals(CharSequence cs) {
        if (count != cs.length())
            return false;
        if (equals(cs)) {
        	return true;
        }
        for (int i = 0; i < count; i++) {
        	if (this.getChar(i) != cs.charAt(i)) {
        		return false;
        	}
        }
        return true;
    }

    public boolean equalsIgnoreCase(QuickString anotherString) {
    	return toString().equalsIgnoreCase(anotherString.toString());
    }

    public int compareTo(QuickString other) {
    	int result;
    	int i1 = 0;
    	int i2 = 0;
    	while (i1 < this.count && i2 < other.count) {
    		if ((result = (this.getChar(i1++) - other.getChar(i2++))) != 0) {
    			return result;
    		}
    	}
    	if (i1 < this.count) {
    		return 1;
    	}
    	if (i2 < other.count) {
    		return -1;
    	}
    	return 0;
	}

    public static final Comparator<QuickString> CASE_INSENSITIVE_ORDER = new CaseInsensitiveComparator();

	private static class CaseInsensitiveComparator implements
			Comparator<QuickString>, java.io.Serializable {
		
		private static final long serialVersionUID = 8575799808933029326L;

		public int compare(QuickString s1, QuickString s2) {
			int i1 = 0;
	    	int i2 = 0;
	    	while (i1 < s1.count && i2 < s2.count) {
	    		char c1 = s1.getChar(i1++);
	    		char c2 = s2.getChar(i2++);
	    		if (c1 != c2) {
					c1 = Character.toUpperCase(c1);
					c2 = Character.toUpperCase(c2);
					if (c1 != c2) {
						c1 = Character.toLowerCase(c1);
						c2 = Character.toLowerCase(c2);
						if (c1 != c2) {
							return c1 - c2;
						}
					}
				}
	    	}
	    	if (i1 < s1.count) {
	    		return 1;
	    	}
	    	if (i2 < s2.count) {
	    		return -1;
	    	}
	    	return 0;
		}
	}

    public int compareToIgnoreCase(QuickString str) {
        return CASE_INSENSITIVE_ORDER.compare(this, str);
    }

    public boolean regionMatches(int toffset, QuickString other, int ooffset, int len) {
    	return toString().regionMatches(toffset, other.toString(), ooffset, len);
	}

    public boolean regionMatches(boolean ignoreCase, int toffset,
                           QuickString other, int ooffset, int len) {
    	return toString().regionMatches(ignoreCase, toffset, other.toString(), ooffset, len);
    }

    public boolean startsWith(QuickString prefix, int toffset) {
    	if ((toffset < 0) || (toffset > count - prefix.count)) {
			return false;
		}
    	return substring(toffset, toffset + prefix.count).equals(prefix);
	}

    public boolean startsWith(QuickString prefix) {
		return startsWith(prefix, 0);
	}

    public boolean endsWith(QuickString suffix) {
		return startsWith(suffix, count - suffix.count);
	}
    
    @Override
    public int hashCode() {
        if (count != 0 && hash == 0) {
            int h = left.hashCode() * pow(31, value.length()) + value.hashCode();
            h = h * pow(31, right.count) + right.hashCode();
            hash = h;
        }
        return hash;
    }
    
    private int pow(int a, int b) {
    	int result = 1;
    	while (b > 0) {
    		if ((b & 1) == 1) {
    			result *= a;
    		}
    		a *= a;
    		b /= 2;
    	}
    	return result;
    }

    public int indexOf(int ch) {
		return indexOf(ch, 0);
	}

    public int indexOf(int ch, int fromIndex) {
		if (ch < Character.MIN_SUPPLEMENTARY_CODE_POINT) {
			return findChar((char) ch, fromIndex);
		}
		if (ch <= Character.MAX_CODE_POINT) {
			char[] surrogates = Character.toChars(ch);
			int i = findChar(surrogates[0], fromIndex);
			int j = findChar(surrogates[1], fromIndex);
			while (i >= 0 && j >= 0 && i + 1 != j) {
				if (i + 1 < j) {
					i = findChar(surrogates[0], i + 1);
				} else {
					j = findChar(surrogates[1], j + 1);
				}
			}
			if (i != -1 && j != -1) {
				return i;
			}
		}
		return -1;
	}

    private int findChar(char ch, int fromIndex) {
		if (count == 0 || fromIndex >= count) {
			return -1;
		}
		int result = left.findChar(ch, fromIndex);
		if (result != -1) {
			return result;
		}
		for (int i = Math.max(0, fromIndex - left.count); i < value.length(); i++) {
			if (value.charAt(i) == ch) {
				return left.count + i;
			}
		}
		result = right.findChar(ch, fromIndex - left.count - value.length());
		if (result != -1) {
			return left.count + value.length() + result;
		}
		return -1;
	}

	public int lastIndexOf(int ch) {
		return lastIndexOf(ch, count - 1);
	}

    public int lastIndexOf(int ch, int fromIndex) {
    	if (ch < Character.MIN_SUPPLEMENTARY_CODE_POINT) {
			return findLastChar((char) ch, fromIndex);
		}
		if (ch <= Character.MAX_CODE_POINT) {
			char[] surrogates = Character.toChars(ch);
			int i = findLastChar(surrogates[0], fromIndex);
			int j = findLastChar(surrogates[1], fromIndex);
			while (i >= 0 && j >= 0 && i + 1 != j) {
				if (i + 1 > j) {
					i = findLastChar(surrogates[0], i - 1);
				} else {
					j = findLastChar(surrogates[1], j - 1);
				}
			}
			if (i != -1 && j != -1) {
				return i;
			}
		}
		return -1;
	}
    
    private int findLastChar(char ch, int fromIndex) {
		if (count == 0 || fromIndex < 0) {
			return -1;
		}
		int result = right.findLastChar(ch, fromIndex - left.count - value.length());
		if (result != -1) {
			return left.count + value.length() + result;
		}
		for (int i = Math.min(value.length() - 1, fromIndex - left.count); i >= 0; i--) {
			if (value.charAt(i) == ch) {
				return i;
			}
		}
		result = left.findLastChar(ch, fromIndex);
		if (result != -1) {
			return result;
		}
		return -1;
	}

    public int indexOf(QuickString str) {
		return indexOf(str, 0);
	}

    public int indexOf(QuickString str, int fromIndex) {
    	if (fromIndex < 0) {
    		fromIndex = 0;
    	}
    	int result = substring(fromIndex).toString().indexOf(str.toString());
    	if (result != -1) {
    		return fromIndex + result;
    	}
    	return -1;
    }

    public int lastIndexOf(QuickString str) {
    	return lastIndexOf(str, count);
    }

    public int lastIndexOf(QuickString str, int fromIndex) {
    	if (fromIndex >= count) {
    		fromIndex = count - 1;
    	}
    	return substring(0, fromIndex + 1).toString().lastIndexOf(str.toString());
    }

    private QuickString[] split(int index) {
    	if (count == 0) {
    		return new QuickString[] {EMPTY, EMPTY};
    	}
    	if (index <= left.count) {
    		QuickString[] tmp = left.split(index);
    		return new QuickString[] {tmp[0], new QuickString(value, tmp[1], right)};
    	} else if (index >= left.count + value.length()) {
    		QuickString[] tmp = right.split(index - left.count - value.length());
    		return new QuickString[] {new QuickString(value, left, tmp[0]), tmp[1]};
    	} else {
    		QuickString l = new QuickString(value.substring(0, index - left.count), left, EMPTY);
    		QuickString r = new QuickString(value.substring(index - left.count), EMPTY, right);
    		return new QuickString[] {l, r};
    	}
    }
    
    public QuickString substring(int beginIndex) {
    	if (beginIndex < 0) {
    	    throw new StringIndexOutOfBoundsException(beginIndex);
    	}
    	if (beginIndex > count) {
    	    throw new StringIndexOutOfBoundsException(count - beginIndex);
    	}
    	return split(beginIndex)[1];
    }

    public QuickString substring(int beginIndex, int endIndex) {
		if (beginIndex < 0) {
			throw new StringIndexOutOfBoundsException(beginIndex);
		}
		if (endIndex > count) {
			throw new StringIndexOutOfBoundsException(endIndex);
		}
		if (beginIndex > endIndex) {
			throw new StringIndexOutOfBoundsException(endIndex - beginIndex);
		}
		return ((beginIndex == 0) && (endIndex == count)) ? this : split(endIndex)[0].split(beginIndex)[1];
	}

    public CharSequence subSequence(int beginIndex, int endIndex) {
        return this.substring(beginIndex, endIndex);
    }

    public QuickString concat(QuickString other) {
    	if (this.count == 0) {
    		return other;
    	}
    	if (other.count == 0) {
    		return this;
    	}
    	if (((rnd += 366239) & 2147483647) % (this.count + other.count) < this.count) {
    		return new QuickString(this.value, this.left, this.right.concat(other));
    	} else {
    		return new QuickString(other.value, this.concat(other.left), other.right);
    	}
    }

    public QuickString replace(char oldChar, char newChar) {
    	if (count == 0) {
    		return this;
    	} else {
	    	QuickString left = this.left.replace(oldChar, newChar);
	    	QuickString right = this.right.replace(oldChar, newChar);
	    	String mid = value.replace(oldChar, newChar);
	    	return new QuickString(mid, left, right);
    	}
	}

    public boolean matches(QuickString regex) {
        return Pattern.matches(regex.toString(), this.toString());
    }

    public boolean contains(CharSequence s) {
        return toString().contains(s);
    }

    public QuickString replaceFirst(QuickString regex, QuickString replacement) {
		return new QuickString(Pattern.compile(regex.toString()).matcher(this.toString()).replaceFirst(replacement.toString()));
	}

    public QuickString replaceAll(QuickString regex, QuickString replacement) {
		return new QuickString(Pattern.compile(regex.toString()).matcher(this.toString()).replaceAll(replacement.toString()));
	}

    public QuickString replace(CharSequence target, CharSequence replacement) {
		return new QuickString(Pattern.compile(target.toString(), Pattern.LITERAL)
				.matcher(this).replaceAll(
						Matcher.quoteReplacement(replacement.toString())));
	}

    public QuickString[] split(QuickString regex, int limit) {
		String[] tmp = Pattern.compile(regex.toString()).split(this, limit);
		QuickString[] result = new QuickString[tmp.length];
		for (int i = 0; i < result.length; i++) {
			result[i] = new QuickString(tmp[i]);
		}
		return result;
	}

    public QuickString[] split(QuickString regex) {
        return split(regex, 0);
    }

    public QuickString toLowerCase(Locale locale) {
    	return new QuickString(toString().toLowerCase(locale));
	}

    public QuickString toLowerCase() {
        return toLowerCase(Locale.getDefault());
    }

    public QuickString toUpperCase(Locale locale) {
    	return new QuickString(toString().toUpperCase(locale));
	}

    public QuickString toUpperCase() {
        return toUpperCase(Locale.getDefault());
    }

    public QuickString trim() {
    	return new QuickString(toString().trim());
	}

    public String toString() {
    	return new String(this.toCharArray());
    }

    public char[] toCharArray() {
		char result[] = new char[count];
		getChars(0, count, result, 0);
		return result;
	}

    public static QuickString format(QuickString format, Object ... args) {
    	return new QuickString(new Formatter().format(format.toString(), args).toString());
    }

    public static QuickString format(Locale l, QuickString format, Object ... args) {
    	return new QuickString(new Formatter(l).format(format.toString(), args).toString());
    }

    public static QuickString valueOf(Object obj) {
    	return new QuickString((obj == null) ? "null" : obj.toString());
    }

    public static QuickString valueOf(char data[]) {
    	return new QuickString(data);
    }

    public static QuickString valueOf(char data[], int offset, int count) {
    	return new QuickString(data, offset, count);
    }

    public static QuickString copyValueOf(char data[], int offset, int count) {
    	return new QuickString(data, offset, count);
    }

    public static QuickString copyValueOf(char data[]) {
    	return new QuickString(data);
    }

    public static QuickString valueOf(boolean b) {
		return new QuickString(b ? "true" : "false");
	}

    public static QuickString valueOf(char c) {
    	return new QuickString(String.valueOf(c));
    }

    public static QuickString valueOf(int i) {
        return new QuickString(Integer.toString(i, 10));
    }

    public static QuickString valueOf(long l) {
        return new QuickString(Long.toString(l, 10));
    }

    public static QuickString valueOf(float f) {
    	return new QuickString(Float.toString(f));
    }

    public static QuickString valueOf(double d) {
    	return new QuickString(Double.toString(d));
    }

    public String intern() {
    	return this.toString().intern();
    }
}
