package ru.ifmo.rain.abdrashitov.java.lang;

public final class QuickStringBuilder implements Appendable, java.io.Serializable,
		CharSequence {

	private static final long serialVersionUID = 5154032399106881429L;
	
	private int log;
	private QuickString quick;
	private StringBuilder buffer;
	
	public QuickStringBuilder() {
		quick = new QuickString();
		buffer = new StringBuilder();
		log = 8;
	}

	public QuickStringBuilder(int capacity) {
		quick = new QuickString();
		buffer = new StringBuilder(capacity);
		log = capacity / 2;
	}

	public QuickStringBuilder(QuickString str) {
		this();
		append(str);
	}
	
	public QuickStringBuilder(QuickStringBuilder other) {
		this.quick = other.quick;
		this.buffer = new StringBuilder(other.buffer);
		this.log = other.log;
	}
	
	public QuickStringBuilder(CharSequence seq) {
		this();
		append(seq);
	}

	public QuickStringBuilder append(Object obj) {
		return append(obj.toString());
	}
	
	public QuickStringBuilder append(String str) {
		return append(new QuickString(str));
	}

	public QuickStringBuilder append(QuickString str) {
		if (buffer.length() != 0) {
			clearBuffer();
		}
		quick = quick.concat(str);
		return this;
	}
	
	private void checkSize() {
		if (buffer.length() > log) {
			clearBuffer();
		}
	}

	private void clearBuffer() {
		quick = quick.concat(new QuickString(buffer.toString()));
		log = 8 + log(quick.length());
		buffer = new StringBuilder(log + 16);
	}
	
	private int log(int x) {
		int result = 0;
		while (x > 1) {
			result++;
			x /= 2;
		}
		return result;
	}

	public QuickStringBuilder append(CharSequence s) {
		if (s instanceof String) {
			return append((String) s);
		}
		buffer.append(s);
		checkSize();
		return this;
	}

	public QuickStringBuilder append(CharSequence s, int start, int end) {
		buffer.append(s, start, end);
		checkSize();
		return this;
	}

	public QuickStringBuilder append(char str[]) {
		buffer.append(str);
		checkSize();
		return this;
	}

	public QuickStringBuilder append(char str[], int offset, int len) {
		buffer.append(str, offset, len);
		checkSize();
		return this;
	}

	public QuickStringBuilder append(boolean b) {
		buffer.append(b);
		checkSize();
		return this;
	}

	public QuickStringBuilder append(char c) {
		buffer.append(c);
		checkSize();
		return this;
	}

	public QuickStringBuilder append(int i) {
		buffer.append(i);
		checkSize();
		return this;
	}

	public QuickStringBuilder append(long lng) {
		buffer.append(lng);
		checkSize();
		return this;
	}

	public QuickStringBuilder append(float f) {
		buffer.append(f);
		checkSize();
		return this;
	}

	public QuickStringBuilder append(double d) {
		buffer.append(d);
		checkSize();
		return this;
	}

	public QuickStringBuilder appendCodePoint(int codePoint) {
		buffer.appendCodePoint(codePoint);
		checkSize();
		return this;
	}

	public QuickStringBuilder delete(int start, int end) {
		clearBuffer();
		quick = quick.substring(0, start).concat(quick.substring(end));
		return this;
	}

	public QuickStringBuilder deleteCharAt(int index) {
		clearBuffer();
		quick = quick.substring(0, index).concat(quick.substring(index + 1));
		return this;
	}

	public QuickStringBuilder replace(int start, int end, QuickString str) {
		clearBuffer();
		quick = quick.substring(0, start).concat(str).concat(quick.substring(end));
		return this;
	}

	public QuickStringBuilder insert(int index, char str[], int offset, int len) {
		clearBuffer();
		quick = quick.substring(0, index).concat(new QuickString(str, offset, len)).concat(quick.substring(index));
		return this;
	}

	public QuickStringBuilder insert(int offset, Object obj) {
		return insert(offset, QuickString.valueOf(obj));
	}

	public QuickStringBuilder insert(int offset, QuickString str) {
		clearBuffer();
		quick = quick.substring(0, offset).concat(str).concat(quick.substring(offset));
		return this;
	}

	public QuickStringBuilder insert(int offset, char str[]) {
		return insert(offset, new QuickString(str));
	}

	public QuickStringBuilder insert(int dstOffset, CharSequence s) {
		if (s == null)
			s = "null";
		if (s instanceof String)
			return insert(dstOffset, new QuickString((String) s));
		return insert(dstOffset, s, 0, s.length());
	}

	public QuickStringBuilder insert(int dstOffset, CharSequence s, int start,
			int end) {
		return insert(dstOffset, new QuickString(s.subSequence(start, end).toString()));
	}

	public QuickStringBuilder insert(int offset, boolean b) {
		return insert(offset, QuickString.valueOf(b));
	}

	public QuickStringBuilder insert(int offset, char c) {
		return insert(offset, QuickString.valueOf(c));
	}

	public QuickStringBuilder insert(int offset, int i) {
		return insert(offset, QuickString.valueOf(i));
	}

	public QuickStringBuilder insert(int offset, long l) {
		return insert(offset, QuickString.valueOf(l));
	}

	public QuickStringBuilder insert(int offset, float f) {
		return insert(offset, QuickString.valueOf(f));
	}

	public QuickStringBuilder insert(int offset, double d) {
		return insert(offset, QuickString.valueOf(d));
	}

	public int indexOf(QuickString str) {
		return indexOf(str, 0);
	}

	public int indexOf(QuickString str, int fromIndex) {
		clearBuffer();
		return quick.indexOf(str, fromIndex);
	}

	public int lastIndexOf(QuickString str) {
		return lastIndexOf(str, length());
	}

	public int lastIndexOf(QuickString str, int fromIndex) {
		clearBuffer();
		return quick.lastIndexOf(str, fromIndex);
	}

	public QuickStringBuilder reverse() {
		buffer = new StringBuilder(toString());
		quick = new QuickString();
		buffer.reverse();
		return this;
	}

	public String toString() {
		if (buffer.length() != 0) {
			clearBuffer();
		}
		return quick.toString();
	}
	
	public QuickString toQuickString() {
		if (buffer.length() != 0) {
			clearBuffer();
		}
		return quick;
	}

	@Override
	public char charAt(int index) {
		if (buffer.length() != 0) {
			clearBuffer();
		}
		return quick.charAt(index);
	}

	@Override
	public int length() {
		return quick.length() + buffer.length();
	}

	@Override
	public CharSequence subSequence(int start, int end) {
		clearBuffer();
		return quick.subSequence(start, end);
	}

}
